from twitter.common import app

from kazoo.client import KazooClient
from twitter.common.zookeeper.group.kazoo_group import KazooGroup


def main():
  import code
  code.interact(local=globals())


app.main()
