
# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: PythonLibrary(BuildFileAddress(FilesystemBuildFile(/Users/jdorlus/dev/GitHub/twitter_commons/src/python/twitter/common/zookeeper/BUILD), zookeeper))

from setuptools import setup

setup(**
{   'description': "Implementations of Twitter's service discovery libraries on top of Kazoo.",
    'extras_require': {   'old': ['zc-zookeeper-static==3.4.4']},
    'install_requires': [   'kazoo==2.2.1',
                            'twitter.common.metrics==0.3.9',
                            'twitter.common.concurrent==0.3.9',
                            'twitter.common.lang==0.3.9',
                            'twitter.common.service-thrift==0.3.9',
                            'thrift',
                            'twitter.common.log==0.3.9',
                            'twitter.common.exceptions==0.3.9',
                            'twitter.common.quantity==0.3.9'],
    'name': 'twitter.common.zookeeper',
    'namespace_packages': ['twitter', 'twitter.common'],
    'package_data': {   },
    'package_dir': {   '': 'src'},
    'packages': [   'twitter',
                    'twitter.common',
                    'twitter.common.zookeeper',
                    'twitter.common.zookeeper.group',
                    'twitter.common.zookeeper.serverset'],
    'version': '0.3.9'}
)
